from dataclasses import asdict, dataclass, field
from datetime import date, datetime, time
from typing import List, Optional

from tonalite.config import Config
from tonalite.core import from_dict

from octo_client import const


@dataclass
class BaseModel:
    @staticmethod
    def _datetime_from_iso_format(datetime_str: Optional[str]) -> Optional[datetime]:
        if datetime_str:
            return datetime.fromisoformat(datetime_str.replace("Z", "+00:00"))
        return None

    def as_dict(self) -> dict:
        """
        Dumps dataclass into dictionary.
        """
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        """
        Populating data class from a dictionary with strict type checking.
        """
        if config is None:
            config = Config(
                strict=strict,
                type_hooks={
                    date: date.fromisoformat,
                    datetime: cls._datetime_from_iso_format,
                },
            )
        return from_dict(cls, data=data, config=config)


@dataclass
class SupplierContact(BaseModel):
    website: Optional[str] = None
    email: Optional[str] = None
    telephone: Optional[str] = None
    address: Optional[str] = None


@dataclass
class Supplier(BaseModel):
    id: str
    name: str
    endpoint: str
    contact: SupplierContact


@dataclass
class Capability(BaseModel):
    id: str
    revision: int
    required: bool


@dataclass
class UnitRestrictions(BaseModel):
    minAge: int
    maxAge: int
    idRequired: bool
    paxCount: int
    accompaniedBy: List[str] = field(default_factory=list)
    minQuantity: Optional[int] = None
    maxQuantity: Optional[int] = None


@dataclass
class Unit(BaseModel):
    id: str
    internalName: str
    type: const.UnitType
    restrictions: UnitRestrictions
    requiredContactFields: List[const.RequiredContactField] = field(default_factory=list)
    reference: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                strict=strict,
                cast=[
                    const.UnitType,
                    const.RequiredContactField,
                ],
            ),
        )


@dataclass
class OptionRestriction(BaseModel):
    minUnits: Optional[int] = None
    maxUnits: Optional[int] = None


@dataclass
class Option(BaseModel):
    id: str
    default: bool
    internalName: str
    cancellationCutoff: str
    cancellationCutoffAmount: int
    cancellationCutoffUnit: const.CancellationCutoffUnit
    restrictions: OptionRestriction
    reference: Optional[str] = None
    availabilityLocalStartTimes: List[str] = field(default_factory=list)
    requiredContactFields: List[const.RequiredContactField] = field(default_factory=list)
    units: List[Unit] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                strict=strict,
                cast=[
                    const.CancellationCutoffUnit,
                    const.RequiredContactField,
                ],
            ),
        )


@dataclass
class Product(BaseModel):
    id: str
    internalName: str
    locale: str
    timeZone: str
    allowFreesale: bool
    instantConfirmation: bool
    instantDelivery: bool
    availabilityRequired: bool
    availabilityType: const.AvailabilityType
    redemptionMethod: const.RedemptionMethod
    options: List[Option]
    deliveryFormats: List[const.DeliveryFormat] = field(default_factory=list)
    deliveryMethods: List[const.DeliveryMethod] = field(default_factory=list)
    reference: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                strict=strict,
                cast=[
                    const.AvailabilityType,
                    const.RedemptionMethod,
                    const.DeliveryFormat,
                    const.DeliveryMethod,
                ],
            ),
        )


@dataclass
class OpeningHours(BaseModel):
    from_: time
    to: time

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        # renaming the keyword
        data["from_"] = data.pop("from")
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(strict=strict, type_hooks={time: time.fromisoformat}),
        )


@dataclass
class AvailabilityCalendarItem(BaseModel):
    localDate: date
    available: bool
    status: const.AvailabilityStatus
    vacancies: Optional[int] = None
    capacity: Optional[int] = None
    openingHours: List[OpeningHours] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                strict=strict,
                type_hooks={
                    date: date.fromisoformat,
                },
                cast=[
                    const.AvailabilityStatus,
                ],
            ),
        )


@dataclass
class Availability(BaseModel):
    id: str
    localDateTimeStart: datetime
    localDateTimeEnd: datetime
    allDay: bool
    available: bool
    status: const.AvailabilityStatus
    utcCutoffAt: datetime
    openingHours: List[OpeningHours] = field(default_factory=list)
    vacancies: Optional[int] = None
    capacity: Optional[int] = None
    maxUnits: Optional[int] = None

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                strict=strict,
                type_hooks={
                    datetime: cls._datetime_from_iso_format,
                },
                cast=[
                    const.AvailabilityStatus,
                ],
            ),
        )


@dataclass
class UnitItem(BaseModel):
    unitId: str
    uuid: Optional[str] = None


@dataclass
class BookingContact(BaseModel):
    locales: List[str] = field(default_factory=list)
    fullName: Optional[str] = None
    firstName: Optional[str] = None
    lastName: Optional[str] = None
    emailAddress: Optional[str] = None
    phoneNumber: Optional[str] = None
    postalCode: Optional[str] = None
    country: Optional[str] = None
    notes: Optional[str] = None


@dataclass
class ConfirmationUnitItem(BaseModel):
    unitId: str
    uuid: Optional[str] = None
    resellerReference: Optional[str] = None
    contact: Optional[BookingContact] = None


@dataclass
class UnitQuantity(BaseModel):
    id: str
    quantity: int


@dataclass
class DeliveryOption(BaseModel):
    deliveryFormat: const.DeliveryFormat
    deliveryValue: str

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                strict=strict,
                cast=[
                    const.DeliveryFormat,
                ],
            ),
        )


@dataclass
class Ticket(BaseModel):
    redemptionMethod: const.RedemptionMethod
    utcRedeemedAt: Optional[datetime] = None
    deliveryOptions: List[DeliveryOption] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                strict=strict,
                type_hooks={
                    datetime: cls._datetime_from_iso_format,
                },
                cast=[
                    const.RedemptionMethod,
                ],
            ),
        )


@dataclass
class BookingUnitItem(BaseModel):
    uuid: str
    unitId: str
    unit: Unit
    status: const.BookingStatus
    contact: BookingContact
    resellerReference: Optional[str] = None
    supplierReference: Optional[str] = None
    utcRedeemedAt: Optional[datetime] = None
    ticket: Optional[Ticket] = None

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                strict=strict,
                type_hooks={
                    datetime: cls._datetime_from_iso_format,
                },
                cast=[
                    const.BookingStatus,
                ],
            ),
        )


@dataclass
class CancellationRequest(BaseModel):
    reason: str
    reasonDetails: str
    status: str
    refund: str
    utcRequestedAt: datetime
    utcHoldExpiration: Optional[datetime] = None
    utcConfirmedAt: Optional[datetime] = None
    utcResolvedAt: Optional[datetime] = None


@dataclass
class Cancellation(BaseModel):
    refund: const.Refund
    reason: Optional[str] = None
    utcCancelledAt = datetime

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                strict=strict,
                type_hooks={
                    datetime: cls._datetime_from_iso_format,
                },
                cast=[
                    const.Refund,
                ],
            ),
        )


@dataclass
class BookingAvailability(BaseModel):
    id: str
    localDateTimeStart: datetime
    localDateTimeEnd: datetime
    allDay: bool
    openingHours: List[OpeningHours] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                strict=strict,
                type_hooks={
                    datetime: cls._datetime_from_iso_format,
                },
            ),
        )


@dataclass
class Booking(BaseModel):
    id: str
    uuid: str
    testMode: bool
    status: const.BookingStatus
    productId: str
    product: Product
    optionId: str
    option: Option
    cancellable: bool
    freesale: bool
    availabilityId: str
    availability: BookingAvailability
    contact: BookingContact
    notes: Optional[str] = None
    voucher: Optional[Ticket] = None
    resellerReference: Optional[str] = None
    supplierReference: Optional[str] = None
    utcCreatedAt: Optional[datetime] = None
    utcUpdatedAt: Optional[datetime] = None
    utcExpiresAt: Optional[datetime] = None
    utcRedeemedAt: Optional[datetime] = None
    utcConfirmedAt: Optional[datetime] = None
    cancellation: Optional[Cancellation] = None
    deliveryMethods: List[const.DeliveryMethod] = field(default_factory=list)
    unitItems: List[BookingUnitItem] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict, config: Optional[Config] = None, strict: bool = False):
        return from_dict(
            data_class=cls,
            data=data,
            config=Config(
                strict=strict,
                type_hooks={
                    datetime: cls._datetime_from_iso_format,
                },
                cast=[
                    const.BookingStatus,
                    const.DeliveryMethod,
                ],
            ),
        )



from enum import Enum


class EnumWithMissing(Enum):
    """An Enum class that overrides the `_missing_()` method to return the class' OTHER member."""
    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value == value:
                return member
        return cls.OTHER


class UnitType(EnumWithMissing):
    ADULT = "ADULT"
    YOUTH = "YOUTH"
    CHILD = "CHILD"
    INFANT = "INFANT"
    FAMILY = "FAMILY"
    SENIOR = "SENIOR"
    STUDENT = "STUDENT"
    MILITARY = "MILITARY"
    OTHER = "OTHER"


class RequiredContactField(Enum):
    firstName = "firstName"
    lastName = "lastName"
    fullName = "fullName"  # ZAUI ONLY
    emailAddress = "emailAddress"
    phoneNumber = "phoneNumber"
    country = "country"
    notes = "notes"
    locales = "locales"


class CancellationCutoffUnit(Enum):
    hour = "hour"
    minute = "minute"
    day = "day"


class AvailabilityType(Enum):
    START_TIME = "START_TIME"
    OPENING_HOURS = "OPENING_HOURS"


class DeliveryFormat(EnumWithMissing):
    PDF_URL = "PDF_URL"
    QRCODE = "QRCODE"
    CODE128 = "CODE128"
    OTHER = "OTHER"


class DeliveryMethod(Enum):
    VOUCHER = "VOUCHER"
    TICKET = "TICKET"


class RedemptionMethod(Enum):
    DIGITAL = "DIGITAL"
    PRINT = "PRINT"
    MANIFEST = "MANIFEST"


class AvailabilityStatus(Enum):
    AVAILABLE = "AVAILABLE"
    FREESALE = "FREESALE"
    SOLD_OUT = "SOLD_OUT"
    LIMITED = "LIMITED"
    CLOSED = "CLOSED"


class BookingStatus(Enum):
    ON_HOLD = "ON_HOLD"
    CONFIRMED = "CONFIRMED"
    EXPIRED = "EXPIRED"
    CANCELLED = "CANCELLED"
    REDEEMED = "REDEEMED"
    PENDING = "PENDING"
    REJECTED = "REJECTED"


class Refund(Enum):
    FULL = "FULL"
    PARTIAL = "PARTIAL"
    NONE = "NONE"


class CancelReason(Enum):
    # is the most common and indicates that the customer requested the cancellation
    CUSTOMER = "CUSTOMER"

    # indicates that the supplier requested the cancellation
    # (possibly due to bad weather or other unexpected circumstances)
    SUPPLIER = "SUPPLIER"

    # indicates that the booking cancellation is being requested by the Reseller
    # because it has been determined the booking was fraudulent
    FRAUD = "FRAUD"

    # indicates that the cancellation reason does not fall into one of these categories.
    # This SHOULD be used only in rare circumstances
    OTHER = "OTHER"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['octo_client']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.20.0,<3', 'tonalite>=1.7.1,<2']

setup_kwargs = {
    'name': 'octo-api-client',
    'version': '1.1.5',
    'description': 'HTTP client for OCTo (Open Connection for Tourism) APIs.',
    'long_description': "# OCTO API client\n\nPython HTTP client for OCTO (Open Connection for Tourism) APIs.\n\nMore info at [octospec.com](https://octospec.com/)\n\nAPI Specification: https://docs.octo.travel/docs/octo/r6gduoa5ah5ne-octo-api\n\n## Installation\n\n    pip install octo-api-client\n\n## Requirements\n\n* Python v3.7+\n\n## Development\n\n### Getting started\n\n    $ pip install poetry\n    $ poetry install\n\n### Running tests and linters\n\nTo run linters:\n\n    $ poetry run ruff octo_client\n    $ poetry run mypy octo_client\n\nTo run tests:\n\n    $ poetry run pytest\n\n\n## Usage\n\n```\nfrom octo_client import OctoClient\n\nclient = OctoClient('https://octo-api.mysupplier.com', 'MY-SECRET_TOKEN')\nclient.get_suppliers()\n```\n",
    'author': 'Tiqets',
    'author_email': 'connections@tiqets.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
from datetime import date
from typing import Any, Callable, Dict, List, Optional, Union

import requests

from octo_client import exceptions, models

logger = logging.getLogger("octo_client")
logger.setLevel(logging.INFO)


class OctoClient(object):
    """
    HTTP client for OCTo (Open Connection for Tourism) APIs.
    """

    def __init__(
        self,
        url: str,
        token: str,
        custom_logger: Optional[logging.Logger] = None,
        log_size_limit: Optional[int] = None,
        requests_loglevel: int = logging.DEBUG,
        language: str = "en",
        strict: bool = False,
    ) -> None:
        """
        Args:
            url (str): URL under which the OCTO interface is available
            token (str): secret bearer token for the authorization
            custom_logger (Logger): custom logger which the client will use for logging
            log_size_limit (int): max limit of the log size above which the log will be truncated
            requests_loglevel (int): default log level that will be used to log requests and
                                     responses
            language (str): language that will be used in the Accept-Language header of each request
            strict (bool): in the strict mode client will raise an error if the response will
                           contain any additional data outside of the data model provided
                           by the specification
        """
        self.url = url.rstrip("/")
        self.token = token
        self.logger = custom_logger or logger
        self.supplier_url_map: Dict[str, str] = {}
        self.requests_loglevel = requests_loglevel
        self.log_responses = False
        self.log_requests = False
        self.log_size_limit = log_size_limit
        self.language = language
        self.strict = strict

    @staticmethod
    def _raise_for_status(status_code: int, response_text: str) -> None:
        CODE_EXCEPTION_MAP = {
            400: exceptions.InvalidRequest,
            403: exceptions.Unauthorized,
            404: exceptions.ApiError,
            500: exceptions.ApiError,
        }
        if status_code in CODE_EXCEPTION_MAP:
            raise CODE_EXCEPTION_MAP[status_code](response_text)

    def _build_endpoint_url_for_request(self, supplier_id: str, path: str) -> str:
        """Builds the endpoint's URL for making requests to a given supplier.

        Args:
            supplier_id: builds the URL for a supplier with this ID.
            path: use this path to build the URL.

        Returns: the full URL.
        Raises:
            - `exceptions.InvalidRequest` if the supplier ID is unknown.

        """

        if supplier_id not in self.supplier_url_map:
            self.get_suppliers()

        try:
            endpoint_url = self.supplier_url_map[supplier_id]
        except KeyError as e:
            raise exceptions.InvalidRequest("Incorrect supplierId") from e

        cleaned_endpoint = endpoint_url.rstrip("/")

        return (
            cleaned_endpoint
            if cleaned_endpoint.endswith(path)
            else f"{cleaned_endpoint}/{path}"
        )

    def _make_request(
        self,
        http_method: Callable,
        path: str,
        supplier_id=None,
        json: Optional[Dict] = None,
        params=None,
        headers: Optional[Dict] = None,
    ):
        if headers is None:
            headers = {}

        if params is None:
            params = {}

        full_url: str = (
            self._build_endpoint_url_for_request(str(supplier_id), path)
            if supplier_id
            else f"{self.url}/{path}"
        )

        request_log_data: dict = {"json": json, "params": params}

        self.logger.log(
            self.requests_loglevel,
            "Sending request to %s (%s)",
            full_url,
            http_method.__name__.upper(),
            extra={"request": self._filter_request_log_data(len(str(request_log_data)), request_log_data)},
        )
        base_headers = self._get_headers()
        headers = {**base_headers, **headers}
        response = http_method(
            full_url,
            params=params,
            json=json,
            headers=headers,
        )
        self._raise_for_status(response.status_code, response.text)
        try:
            response_json = response.json()
        except Exception as exc:
            self.logger.log(
                self.requests_loglevel,
                "Received non-JSON response",
                extra={"response": self._filter_log_data(len(response.text), response.text)},
            )
            raise exceptions.ApiError("Non-JSON response") from exc
        self.logger.log(
            self.requests_loglevel,
            "Got response from %s (%s)",
            full_url,
            http_method.__name__.upper(),
            extra={"response": self._filter_log_data(len(response.text), response_json)},
        )
        return response_json

    def _filter_request_log_data(
            self, data_length: int, content: Union[str, dict]
    ) -> Optional[Union[str, dict]]:
        if self.log_requests:
            if self.log_size_limit and data_length > self.log_size_limit:
                return "TRUNCATED"
            return content
        return None

    def _filter_log_data(
            self, response_length: int, response_content: Union[str, dict]
    ) -> Optional[Union[str, dict]]:
        if self.log_responses:
            if self.log_size_limit and response_length > self.log_size_limit:
                return "TRUNCATED"
            return response_content
        return None

    def _get_headers(self) -> Dict[str, str]:
        return {
            "Authorization": f"Bearer {self.token}",
            "Accept-Language": self.language,
        }

    def _http_get(self, path: str, supplier_id: Optional[str] = None, params=None, headers: Optional[Dict] = None):
        return self._make_request(requests.get, path, supplier_id=supplier_id, params=params, headers=headers)

    def _http_post(self, path: str, json: dict, supplier_id: str, params=None, headers: Optional[Dict] = None):
        if headers is None:
            headers = {}
        return self._make_request(
            requests.post,
            path,
            json=json,
            supplier_id=supplier_id,
            params=params,
            headers={"Content-Type": "application/json", **headers},
        )

    def _http_patch(self, path: str, json: dict, supplier_id: str, params=None, headers: Optional[Dict] = None):
        if headers is None:
            headers = {}
        return self._make_request(
            requests.patch,
            path,
            json=json,
            supplier_id=supplier_id,
            params=params,
            headers={"Content-Type": "application/json", **headers},
        )

    def _http_delete(self, path: str, supplier_id: str, json: dict, params=None, headers: Optional[Dict] = None):
        if headers is None:
            headers = {}
        return self._make_request(
            requests.delete,
            path,
            supplier_id=supplier_id,
            json=json,
            params=params,
            headers={"Content-Type": "application/json", **headers},
        )

    def get_supplier(self, supplier_id: str, headers: Optional[Dict] = None) -> models.Supplier:
        response = self._http_get(f"suppliers/{supplier_id}", supplier_id=supplier_id, headers=headers)
        try:
            supplier = models.Supplier.from_dict(response, strict=self.strict)
        except AttributeError as e:
            raise exceptions.ApiError(response) from e
        return supplier

    def get_suppliers(self, headers: Optional[Dict] = None) -> List[models.Supplier]:
        response = self._http_get("suppliers", headers=headers)
        try:
            suppliers = [
                models.Supplier.from_dict(supplier, strict=self.strict) for supplier in response
            ]
        except AttributeError as e:
            raise exceptions.ApiError(response) from e
        self.logger.info("Found %s suppliers", len(suppliers), extra={"suppliers": response})
        self.supplier_url_map = {supplier.id: supplier.endpoint for supplier in suppliers}
        return suppliers

    def get_products(self, supplier_id: str, headers: Optional[Dict] = None) -> List[models.Product]:
        response = self._http_get("products", supplier_id=supplier_id, headers=headers)
        products = [models.Product.from_dict(product, strict=self.strict) for product in response]
        self.logger.info("Found %s products", len(products), extra={"products": products})
        return products

    def get_product(self, supplier_id: str, product_id: str, headers: Optional[Dict] = None) -> models.Product:
        response = self._http_get(f"products/{product_id}", supplier_id=supplier_id, headers=headers)
        return models.Product.from_dict(response, strict=self.strict)

    def availability_check(
        self,
        supplier_id: str,
        product_id: str,
        option_id: str,
        units: Optional[List[models.UnitQuantity]] = None,
        local_date_start: Optional[date] = None,
        local_date_end: Optional[date] = None,
        local_date: Optional[date] = None,
        availability_ids: Optional[List[str]] = None,
        headers: Optional[Dict] = None
    ) -> List[models.Availability]:
        payload: Dict[str, Any] = {
            "productId": product_id,
            "optionId": option_id,
        }
        if any([local_date_start, local_date_end]) and not all([local_date_start, local_date_end]):
            raise ValueError("local_date_start and local_date_end needs to be used together")
        if local_date_start:
            payload["localDateStart"] = local_date_start.isoformat()
        if local_date_end:
            payload["localDateEnd"] = local_date_end.isoformat()
        if local_date:
            payload["localDate"] = local_date.isoformat()
        if availability_ids:
            payload["availabilityIds"] = availability_ids
        if units:
            payload["units"] = [unit.as_dict() for unit in units]

        response = self._http_post("availability", supplier_id=supplier_id, json=payload, headers=headers)
        detailed_availability = [
            models.Availability.from_dict(availability, strict=self.strict)
            for availability in response
        ]
        self.logger.info("Found %s items", len(detailed_availability))
        return detailed_availability

    def get_calendar(
        self,
        supplier_id: str,
        product_id: str,
        option_id: str,
        local_date_start: date,
        local_date_end: date,
        units: Optional[List[models.UnitQuantity]] = None,
        headers: Optional[Dict] = None,
    ) -> List[models.AvailabilityCalendarItem]:
        """This method retrieve the availability calendar for a range of dates.

        Args:
            supplier_id: retrieve calendar data for the supplier with this ID.
            product_id: retrieve calendar data for a product with this ID.
            option_id: retrieve calendar data for the option with this ID.
            local_date_start: retrieve calendar data from this date onwards.
            local_date_end: retrieve calendar data until this date.
            units: a list of units. Each unit requires:
                - id
                - quantity
            headers: optional HTTP headers to send in the request.

        Returns: a list of availability objects; one object per each day in the range of dates.
        """

        payload: Dict[str, Any] = {
            "productId": product_id,
            "optionId": option_id,
        }
        if local_date_start:
            payload["localDateStart"] = local_date_start.isoformat()
        if local_date_end:
            payload["localDateEnd"] = local_date_end.isoformat()
        if units:
            payload["units"] = [unit.as_dict() for unit in units]

        response = self._http_post("availability/calendar", supplier_id=supplier_id, json=payload, headers=headers)
        daily_availability = [
            models.AvailabilityCalendarItem.from_dict(availability, strict=self.strict)
            for availability in response
        ]
        self.logger.info("Found %s days", len(daily_availability))
        return daily_availability

    def booking_reservation(
        self,
        supplier_id: str,
        uuid: str,
        product_id: str,
        option_id: str,
        availability_id: str,
        unit_items: List[models.UnitItem],
        expiration_minutes: Optional[int] = None,
        notes: Optional[str] = None,
        headers: Optional[Dict] = None,
    ) -> models.Booking:
        payload: Dict[str, Any] = {
            "uuid": uuid,
            "productId": product_id,
            "optionId": option_id,
            "availabilityId": availability_id,
            "unitItems": [unit.as_dict() for unit in unit_items],
        }
        if expiration_minutes:
            payload["expirationMinutes"] = expiration_minutes
        if notes:
            payload["notes"] = notes
        response = self._http_post(
            "bookings",
            supplier_id=supplier_id,
            json=payload,
            headers=headers,
        )
        self.logger.info("Booking created", extra={"booking": response})
        return models.Booking.from_dict(response)

    def list_bookings(
        self,
        supplier_id: str,
        reseller_reference: Optional[str] = None,
        supplier_reference: Optional[str] = None,
        local_date: Optional[date] = None,
        local_date_start: Optional[date] = None,
        local_date_end: Optional[date] = None,
        headers: Optional[Dict] = None,
    ) -> List[models.Booking]:
        if not any(
            [reseller_reference, supplier_reference, local_date, local_date_start or local_date_end]
        ):
            raise ValueError("One of the query parameters has to be provided.")

        if any([local_date_start, local_date_end]) and not all([local_date_start, local_date_end]):
            raise ValueError("local_date_start and local_date_end needs to be used together")

        params: Dict[str, str] = {}
        if reseller_reference:
            params["resellerReference"] = reseller_reference
        elif supplier_reference:
            params["supplierReference"] = supplier_reference
        elif local_date:
            params["localDate"] = local_date.isoformat()
        elif local_date_start and local_date_end:
            params["localDateStart"] = local_date_start.isoformat()
            params["localDateEnd"] = local_date_end.isoformat()

        response = self._http_get("bookings", supplier_id=supplier_id, params=params, headers=headers)
        return [models.Booking.from_dict(booking) for booking in response]

    def get_booking(self, supplier_id: str, uuid: str, headers: Optional[Dict] = None,) -> models.Booking:
        response = self._http_get(f"bookings/{uuid}", supplier_id=supplier_id, headers=headers)
        return models.Booking.from_dict(response)

    def booking_confirmation(
        self,
        supplier_id: str,
        uuid: str,
        email_receipt: bool = False,
        reseller_reference: Optional[str] = None,
        contact_full_name: Optional[str] = None,
        contact_first_name: Optional[str] = None,
        contact_last_name: Optional[str] = None,
        contact_email_address: Optional[str] = None,
        contact_phone_number: Optional[str] = None,
        contact_locales: Optional[List[str]] = None,
        contact_postal_code: Optional[str] = None,
        contact_country: Optional[str] = None,
        contact_notes: Optional[str] = None,
        unit_items: Optional[List[models.ConfirmationUnitItem]] = None,
        headers: Optional[Dict] = None,
    ) -> models.Booking:
        payload: Dict[str, Any] = {}
        if email_receipt:
            payload["emailReceipt"] = email_receipt
        if reseller_reference:
            payload["resellerReference"] = reseller_reference
        if any(
            [
                contact_full_name,
                contact_first_name,
                contact_last_name,
                contact_email_address,
                contact_phone_number,
                contact_locales,
                contact_postal_code,
                contact_country,
                contact_notes,
            ]
        ):
            payload["contact"] = models.BookingContact.from_dict(
                {
                    "locales": contact_locales or [],
                    "fullName": contact_full_name,
                    "firstName": contact_first_name,
                    "lastName": contact_last_name,
                    "emailAddress": contact_email_address,
                    "phoneNumber": contact_phone_number,
                    "postalCode": contact_postal_code,
                    "country": contact_country,
                    "notes": contact_notes,
                }
            ).as_dict()
        if unit_items:
            payload["unitItems"] = [unit_item.as_dict() for unit_item in unit_items]

        response = self._http_post(
            f"bookings/{uuid}/confirm", supplier_id=supplier_id, json=payload, headers=headers
        )
        return models.Booking.from_dict(response)

    def extend_reservation(
        self, supplier_id: str, uuid: str, expiration_minutes: int, headers: Optional[Dict] = None,
    ) -> models.Booking:
        payload = {"expirationMinutes": expiration_minutes}
        response = self._http_post(f"bookings/{uuid}/extend", supplier_id=supplier_id, json=payload, headers=headers)
        return models.Booking.from_dict(response)

    def booking_cancellation(
        self,
        supplier_id: str,
        uuid: str,
        reason: Optional[str] = None,
        force: Optional[bool] = False,
        headers: Optional[Dict] = None,
    ) -> models.Booking:
        payload: Dict[str, Any] = {}
        if reason:
            payload["reason"] = reason
        if force:
            payload["force"] = force
        response = self._http_delete(f"bookings/{uuid}", supplier_id=supplier_id, json=payload, headers=headers)
        return models.Booking.from_dict(response)

    def booking_update(
        self,
        supplier_id: str,
        uuid: str,
        product_id: Optional[str] = None,
        option_id: Optional[str] = None,
        availability_id: Optional[str] = None,
        notes: Optional[str] = None,
        email_receipt: bool = False,
        reseller_reference: Optional[str] = None,
        contact_full_name: Optional[str] = None,
        contact_first_name: Optional[str] = None,
        contact_last_name: Optional[str] = None,
        contact_email_address: Optional[str] = None,
        contact_phone_number: Optional[str] = None,
        contact_locales: Optional[List[str]] = None,
        contact_postal_code: Optional[str] = None,
        contact_country: Optional[str] = None,
        contact_notes: Optional[str] = None,
        unit_items: Optional[List[models.ConfirmationUnitItem]] = None,
        headers: Optional[Dict] = None,
    ):
        payload: Dict[str, Any] = {}
        if product_id:
            payload["productId"] = product_id
        if option_id:
            payload["optionId"] = option_id
        if availability_id:
            payload["availabilityId"] = availability_id
        if notes:
            payload["notes"] = notes
        if email_receipt:
            payload["emailReceipt"] = email_receipt
        if reseller_reference:
            payload["resellerReference"] = reseller_reference

        if any(
            [
                contact_full_name,
                contact_first_name,
                contact_last_name,
                contact_email_address,
                contact_phone_number,
                contact_locales,
                contact_postal_code,
                contact_country,
                contact_notes,
            ]
        ):
            payload["contact"] = {}
            if contact_locales:
                payload["contact"]["locales"] = contact_locales or []
            if contact_full_name:
                payload["contact"]["fullName"] = contact_full_name
            if contact_first_name:
                payload["contact"]["firstName"] = contact_first_name
            if contact_last_name:
                payload["contact"]["lastName"] = contact_last_name
            if contact_email_address:
                payload["contact"]["emailAddress"] = contact_email_address
            if contact_phone_number:
                payload["contact"]["phoneNumber"] = contact_phone_number
            if contact_postal_code:
                payload["contact"]["postalCode"] = contact_postal_code
            if contact_country:
                payload["contact"]["country"] = contact_country
            if contact_notes:
                payload["contact"]["notes"] = contact_notes

        if unit_items:
            payload["unitItems"] = [unit_item.as_dict() for unit_item in unit_items]

        response = self._http_patch(f"bookings/{uuid}", supplier_id=supplier_id, json=payload, headers=headers)
        return models.Booking.from_dict(response)



class Unauthorized(Exception):
    """
    The Authorization header was validated but the requestor does not have
    the correct permissions to access the requested resource or perform
    the requested operation.
    """


class ApiError(Exception):
    """
    Unexpected error returned by the supplier API.
    """


class InvalidRequest(Exception):
    """
    Invalid request (e.g. missing required parameters).
    """
